@@ -44,6 +44,19 @@ class Redirect(enum.IntEnum):
         assert obj == Redirect.DEFAULT
         return _DEFAULT_REDIRECTION[(fdesc, bool(pty))]
 
+    @staticmethod
+    def from_literal(literal, is_stderr=False):
+        "Return object with literal values replaced by Redirect constant."
+        # For stderr, the literal `1` indicates that stderr is redirected to
+        # the same place as STDOUT.
+        if is_stderr and isinstance(literal, int) and literal == 1:
+            return Redirect.STDOUT
+
+        try:
+            return _LITERAL_REDIRECT.get(literal, literal)
+        except TypeError:
+            return literal
+
 
 # This table has the default redirections for (src, pty).
 # Sources are stdin, stdout, stderr. Used by Redirect.from_default().
@@ -57,9 +70,42 @@ _DEFAULT_REDIRECTION = {
     (_STDERR, True): Redirect.STDOUT,
 }
 
+_LITERAL_REDIRECT = {
+    None: Redirect.DEVNULL,
+    ...: Redirect.INHERIT,
+    (): Redirect.CAPTURE,
+}
+
 # Used in Command and Pipeline to implement operator overloading.
-STDIN_TYPES = (str, bytes, os.PathLike, bytearray, io.IOBase, int, Redirect)
-STDOUT_TYPES = (str, bytes, os.PathLike, bytearray, io.IOBase, int, Redirect, Logger)
+STDIN_TYPES = (
+    str,
+    bytes,
+    os.PathLike,
+    bytearray,
+    io.IOBase,
+    int,
+    Redirect,
+    asyncio.StreamReader,
+    type(None),
+    type(...),
+    tuple,
+)
+
+STDOUT_TYPES = (
+    str,
+    bytes,
+    os.PathLike,
+    bytearray,
+    io.IOBase,
+    int,
+    Redirect,
+    Logger,
+    asyncio.StreamWriter,
+    type(None),
+    type(...),
+    tuple,
+)
+
 STDOUT_APPEND_TYPES = (str, bytes, os.PathLike)
 
 
@@ -101,7 +147,38 @@ async def write_stream(
         # of closing the pty. At the beginning of a line, we only need to
         # send one EOF. Otherwise, we need to send one EOF to end the
         # partial line and then another EOF to signal we are done.
-        if input_bytes and input_bytes[-1] != "\n":
+        if not input_bytes.endswith(b"\n"):
+            stream.write(eof + eof)
+        else:
+            stream.write(eof)
+        await _drain(stream)
+
+
+@log_method(LOG_DETAIL)
+async def write_reader(
+    reader: asyncio.StreamReader,
+    stream: asyncio.StreamWriter,
+    eof: Optional[bytes] = None,
+):
+    "Copy from reader to writer."
+
+    ends_with_newline = True
+    try:
+        while True:
+            data = await reader.read(_CHUNK_SIZE)
+            if not data:
+                break
+            ends_with_newline = data.endswith(b"\n")
+            stream.write(data)
+            await stream.drain()
+
+    except (BrokenPipeError, ConnectionResetError):
+        pass
+
+    if eof is None:
+        stream.close()
+    elif eof:
+        if not ends_with_newline:
             stream.write(eof + eof)
         else:
             stream.write(eof)
@@ -163,6 +240,20 @@ async def copy_bytearray(source: asyncio.StreamReader, dest: bytearray):
         dest.extend(data)
 
 
+@log_method(LOG_DETAIL)
+async def copy_streamwriter(source: asyncio.StreamReader, dest: asyncio.StreamWriter):
+    "Copy bytes from source stream to dest StreamWriter."
+    while True:
+        data = await source.read(_CHUNK_SIZE)
+        if not data:
+            break
+        dest.write(data)
+        await dest.drain()
+
+    dest.close()
+    await dest.wait_closed()
+
+
 @log_method(LOG_DETAIL)
 async def read_lines(source: asyncio.StreamReader, encoding: Optional[str]):
     "Async iterator over lines in stream."
