@@ -13,13 +13,12 @@ from shellous.harvest import harvest, harvest_results
 from shellous.log import LOG_DETAIL, LOG_ENTER, LOG_EXIT, LOGGER, log_method, log_timer
 from shellous.redirect import Redirect
 from shellous.result import Result, make_result
-from shellous.util import close_fds, verify_dev_fd, wait_pid
+from shellous.util import close_fds, uninterrupted, verify_dev_fd, wait_pid, which
 
 _KILL_TIMEOUT = 3.0
 _CLOSE_TIMEOUT = 0.25
+_UNLAUNCHED_EXIT_CODE = -255
 
-_KILL_EXIT_CODE = -9 if sys.platform != "win32" else 1
-_FLAKY_EXIT_CODE = 255
 _BSD = sys.platform.startswith("freebsd") or sys.platform == "darwin"
 
 AUDIT_EVENT_SUBPROCESS_SPAWN = "byllyfish/shellous.subprocess_spawn"
@@ -41,6 +40,12 @@ def _is_write_mode(cmd):
     return cmd.options.write_mode
 
 
+def _split(encoding):
+    if encoding is None:
+        raise TypeError("when encoding is None, input must be bytes")
+    return encoding.split(maxsplit=1)
+
+
 class _RunOptions:  # pylint: disable=too-many-instance-attributes
     """_RunOptions is context manager to assist in running a command.
 
@@ -75,16 +80,21 @@ class _RunOptions:  # pylint: disable=too-many-instance-attributes
             self._setup_redirects()
             return self
         except Exception as ex:
-            LOGGER.warning("_RunOptions.enter %r ex=%r", self.command.name, ex)
+            if LOG_DETAIL:
+                LOGGER.warning("_RunOptions.enter %r ex=%r", self.command.name, ex)
+            _cleanup(self.command)
             raise
 
     def __exit__(self, _exc_type, exc_value, _exc_tb):
         "Make sure those file descriptors are cleaned up."
         self.close_fds()
         if exc_value:
-            LOGGER.warning(
-                "_RunOptions.exit %r exc_value=%r", self.command.name, exc_value
-            )
+            if LOG_DETAIL:
+                LOGGER.warning(
+                    "_RunOptions.exit %r exc_value=%r",
+                    self.command.name,
+                    exc_value,
+                )
             for subcmd in self.subcmds:
                 _cleanup(subcmd)
 
@@ -150,8 +160,8 @@ class _RunOptions:  # pylint: disable=too-many-instance-attributes
         )
 
         # Set up PTY here. This is the first half. Second half in `Runner`.
-        start_new_session = options.start_new_session
-        preexec_fn = options.preexec_fn
+        start_new_session = options._start_new_session
+        preexec_fn = options._preexec_fn
         if options.pty:
             assert preexec_fn is None
             stdin, stdout, stderr, preexec_fn = self._setup_pty1(
@@ -163,7 +173,6 @@ class _RunOptions:  # pylint: disable=too-many-instance-attributes
             start_new_session = True
 
         self.input_bytes = input_bytes
-        self.args = self.command.args
         self.kwd_args = {
             "stdin": stdin,
             "stdout": stdout,
@@ -171,13 +180,19 @@ class _RunOptions:  # pylint: disable=too-many-instance-attributes
             "env": options.merge_env(),
             "start_new_session": start_new_session,
             "preexec_fn": preexec_fn,
+            "close_fds": options.close_fds,
         }
 
         if options.pass_fds:
+            self.kwd_args["close_fds"] = True
             self.kwd_args["pass_fds"] = options.pass_fds
             if options.pass_fds_close:
                 self.open_fds.extend(options.pass_fds)
 
+        self.args = list(self.command.args)
+        if not os.path.dirname(self.args[0]):
+            self.args[0] = which(self.args[0])
+
     def _setup_input(self, input_, close, encoding):
         "Set up process input."
         assert input_ is not None
@@ -204,10 +219,14 @@ class _RunOptions:  # pylint: disable=too-many-instance-attributes
             stdin = input_
             if close:
                 self.open_fds.append(stdin)
+        elif isinstance(input_, str):
+            input_bytes = input_.encode(*_split(encoding))
+        elif isinstance(input_, (asyncio.StreamReader, io.BytesIO, io.StringIO)):
+            # Shellous-supported input classes.
+            assert stdin == asyncio.subprocess.PIPE
+            assert input_bytes is None
         else:
-            if encoding is None:
-                raise TypeError("when encoding is None, input must be bytes")
-            input_bytes = input_.encode(*encoding.split(maxsplit=1))
+            raise TypeError(f"unsupported input type: {input_!r}")
 
         return stdin, input_bytes
 
@@ -234,13 +253,16 @@ class _RunOptions:  # pylint: disable=too-many-instance-attributes
             stdout = output
             if close:
                 self.open_fds.append(stdout)
-        elif isinstance(output, (io.StringIO, io.BytesIO, bytearray, Logger)):
-            pass
+        elif isinstance(
+            output, (io.StringIO, io.BytesIO, bytearray, Logger, asyncio.StreamWriter)
+        ):
+            # Shellous-supported output classes.
+            assert stdout == asyncio.subprocess.PIPE
         elif isinstance(output, io.IOBase):
             # Client-managed File-like object.
             stdout = output
         else:
-            raise TypeError(f"unsupported type: {output!r}")
+            raise TypeError(f"unsupported output type: {output!r}")
 
         return stdout
 
@@ -287,80 +309,94 @@ class Runner:
     ```
     """
 
+    stdin = None
+    "Process standard input."
+
+    stdout = None
+    "Process standard output."
+
+    stderr = None
+    "Process standard error."
+
     def __init__(self, command):
-        self.options = _RunOptions(command)
-        self.cancelled = False
-        self.proc = None
-        self.tasks = []
-        self.stdin = None
-        self.stdout = None
-        self.stderr = None
+        self._options = _RunOptions(command)
+        self._cancelled = False
+        self._proc = None
+        self._tasks = []
 
     @property
     def name(self):
-        "Return name of process to run."
+        "Return name of process being run."
         return self.command.name
 
     @property
     def command(self):
         "Return the command being run."
-        return self.options.command
+        return self._options.command
 
     @property
     def pid(self):
         "Return the command's process ID."
-        if not self.proc:
+        if not self._proc:
             return None
-        return self.proc.pid
+        return self._proc.pid
+
+    @property
+    def returncode(self):
+        "Process's exit code."
+        if not self._proc:
+            return None
+        return self._proc.returncode
+
+    @property
+    def cancelled(self):
+        "Return True if the command was cancelled."
+        return self._cancelled
 
     def result(self, output_bytes=b""):
         "Check process exit code and raise a ResultError if necessary."
-        if self.proc:
-            code = self.proc.returncode
+        if self._proc:
+            code = self._proc.returncode
         else:
-            # The process was started but cancelled immediately; `self.proc`
-            # was never returned.
-            assert self.cancelled
-            code = _KILL_EXIT_CODE
+            # The process was cancelled before starting.
+            assert self._cancelled
+            code = _UNLAUNCHED_EXIT_CODE
 
         result = Result(
             output_bytes,
             code,
-            self.cancelled,
-            self.options.encoding,
+            self._cancelled,
+            self._options.encoding,
         )
 
-        return make_result(self.command, result, self.cancelled)
+        return make_result(self.command, result, self._cancelled)
 
-    def add_task(self, coro, tag=None):
+    def add_task(self, coro, tag=""):
         "Add a background task."
-        if tag:
-            task_name = f"{self.name}#{tag}"
-        else:
-            task_name = self.name
+        task_name = f"{self.name}#{tag}"
         task = asyncio.create_task(coro, name=task_name)
-        self.tasks.append(task)
+        self._tasks.append(task)
         return task
 
     def _is_bsd_pty(self):
         "Return true if we're running a pty on BSD."
-        return _BSD and self.options.pty_fds
+        return _BSD and self._options.pty_fds
 
     @log_method(LOG_DETAIL)
     async def _wait(self):
         "Normal wait for background I/O tasks and process to finish."
-        assert self.proc
+        assert self._proc
 
         try:
-            if self.tasks:
-                await harvest(*self.tasks, trustee=self)
+            if self._tasks:
+                await harvest(*self._tasks, trustee=self)
             if self._is_bsd_pty():
                 await self._waiter()
 
         except asyncio.CancelledError:
             LOGGER.info("Runner.wait cancelled %r", self)
-            self.cancelled = True
-            self.tasks.clear()  # all tasks were cancelled
+            self._cancelled = True
+            self._tasks.clear()  # all tasks were cancelled
             await self._kill()
 
     @log_method(LOG_DETAIL)
@@ -368,11 +404,11 @@ class Runner:
         "Manually poll `waitpid` until process finishes."
         assert self._is_bsd_pty()
         while True:
-            status = wait_pid(self.proc.pid)
+            status = wait_pid(self._proc.pid)
             if status is not None:
                 # pylint: disable=protected-access
-                self.proc._transport._returncode = status
-                self.proc._transport._proc.returncode = status
+                self._proc._transport._returncode = status
+                self._proc._transport._proc.returncode = status
                 break
 
             await asyncio.sleep(0.025)
@@ -380,26 +416,26 @@ class Runner:
     @log_method(LOG_DETAIL)
     async def _kill(self):
         "Kill process and wait for it to finish."
-        assert self.proc
+        assert self._proc
 
         cancel_timeout = self.command.options.cancel_timeout
         cancel_signal = self.command.options.cancel_signal
 
         try:
             # If not already done, send cancel signal.
-            if self.proc.returncode is None:
+            if self._proc.returncode is None:
                 self._send_signal(cancel_signal)
 
-            if self.tasks:
-                await harvest(*self.tasks, timeout=cancel_timeout, trustee=self)
+            if self._tasks:
+                await harvest(*self._tasks, timeout=cancel_timeout, trustee=self)
 
-            if self.proc.returncode is None:
+            if self._proc.returncode is None:
                 await harvest(self._waiter(), timeout=cancel_timeout, trustee=self)
 
         except (asyncio.CancelledError, asyncio.TimeoutError) as ex:
             LOGGER.warning("Runner.kill %r (ex)=%r", self, ex)
             if _is_cancelled(ex):
-                self.cancelled = True
+                self._cancelled = True
             await self._kill_wait()
 
         except Exception as ex:
@@ -411,51 +447,57 @@ class Runner:
         "Send a signal to the process."
 
         LOGGER.info("Runner.signal %r signal=%r", self, sig)
+        self._audit_callback("signal", signal=sig)
+
         if sig is None:
-            self.proc.kill()
+            self._proc.kill()
         else:
-            self.proc.send_signal(sig)
+            self._proc.send_signal(sig)
 
     @log_method(LOG_DETAIL)
     async def _kill_wait(self):
         "Wait for killed process to exit."
-        assert self.proc
+        assert self._proc
 
         # Check if process is already done.
-        if self.proc.returncode is not None:
+        if self._proc.returncode is not None:
             return
 
         try:
             self._send_signal(None)
             await harvest(self._waiter(), timeout=_KILL_TIMEOUT, trustee=self)
         except asyncio.TimeoutError as ex:
-            LOGGER.error("%r failed to kill process %r", self, self.proc)
-            raise RuntimeError(f"Unable to kill process {self.proc!r}") from ex
+            LOGGER.error("%r failed to kill process %r", self, self._proc)
+            raise RuntimeError(f"Unable to kill process {self._proc!r}") from ex
 
     @log_method(LOG_ENTER, _info=True)
     async def __aenter__(self):
         "Set up redirections and launch subprocess."
+        self._audit_callback("start")
         try:
             return await self._start()
+        except BaseException as ex:
+            self._audit_callback("stop", failure=type(ex).__name__)
+            raise
         finally:
-            if self.cancelled and self.command.options.incomplete_result:
+            if self._cancelled and self.command.options.incomplete_result:
                 # Raises ResultError instead of CancelledError.
                 self.result()
 
     @log_method(LOG_DETAIL)
     async def _start(self):
         "Set up redirections and launch subprocess."
-        assert self.proc is None
-        assert not self.tasks
+        assert self._proc is None
+        assert not self._tasks
 
         try:
             # Set up subprocess arguments and launch subprocess.
-            with self.options as opts:
+            with self._options as opts:
                 await self._subprocess_spawn(opts)
 
-            stdin = self.proc.stdin
-            stdout = self.proc.stdout
-            stderr = self.proc.stderr
+            stdin = self._proc.stdin
+            stdout = self._proc.stdout
+            stderr = self._proc.stderr
 
             # Assign pty streams.
             if opts.pty_fds:
@@ -473,19 +515,13 @@ class Runner:
                 )
 
             if stdin is not None:
-                if opts.input_bytes is not None:
-                    eof = opts.pty_fds.eof if opts.pty_fds else None
-                    self.add_task(
-                        redir.write_stream(opts.input_bytes, stdin, eof),
-                        "stdin",
-                    )
-                    stdin = None
+                stdin = self._setup_input_source(stdin, opts)
 
         except (Exception, asyncio.CancelledError) as ex:
             LOGGER.info("Runner._start %r ex=%r", self, ex)
             if _is_cancelled(ex):
-                self.cancelled = True
-            if self.proc:
+                self._cancelled = True
+            if self._proc:
                 await self._kill()
             raise
 
@@ -503,7 +539,7 @@ class Runner:
 
     @log_method(LOG_DETAIL)
     async def _subprocess_spawn(self, opts):
-        "Start the subprocess and assign to `self.proc`."
+        "Start the subprocess."
 
         # Second half of pty setup.
         if opts.pty_fds:
@@ -511,25 +547,61 @@ class Runner:
             if _BSD:
                 pty_util.patch_child_watcher()
 
-        with log_timer("asyncio.create_subprocess_exec"):
-            # AUDIT: subprocess spawn
-            sys.audit(AUDIT_EVENT_SUBPROCESS_SPAWN, opts.args[0])
-            self.proc = await asyncio.create_subprocess_exec(
-                *opts.args,
-                **opts.kwd_args,
-            )
+        # Check for task cancellation and yield right before exec'ing. If the
+        # current task is already cancelled, this will raise a CancelledError,
+        # and we save ourselves the work of launching and immediately killing
+        # a process.
+        await asyncio.sleep(0)
+        # Launch the subprocess (always completes even if cancelled).
+        await uninterrupted(self._subprocess_exec(opts))
 
         # Launch the process substitution commands (if any).
         for cmd in opts.subcmds:
             self.add_task(cmd.coro(), "procsub")
 
+    @log_method(LOG_DETAIL)
+    async def _subprocess_exec(self, opts):
+        "Start the subprocess and assign to `self.proc`."
+        with log_timer("asyncio.create_subprocess_exec"):
+            sys.audit(AUDIT_EVENT_SUBPROCESS_SPAWN, opts.args[0])
+            self._proc = await asyncio.create_subprocess_exec(
+                *opts.args,
+                **opts.kwd_args,
+            )
+
     @log_method(LOG_DETAIL)
     async def _waiter(self):
         "Run task that waits for process to exit."
         if self._is_bsd_pty():
             await self._wait_pid()
         else:
-            await self.proc.wait()
+            await self._proc.wait()
+
+    def _setup_input_source(self, stream, opts):
+        "Set up a task to read from custom input source."
+        tag = "stdin"
+        eof = opts.pty_fds.eof if opts.pty_fds else None
+
+        if opts.input_bytes is not None:
+            self.add_task(redir.write_stream(opts.input_bytes, stream, eof), tag)
+            return None
+
+        source = opts.command.options.input
+
+        if isinstance(source, asyncio.StreamReader):
+            self.add_task(redir.write_reader(source, stream, eof), tag)
+            return None
+
+        if isinstance(source, io.BytesIO):
+            self.add_task(redir.write_stream(source.getvalue(), stream, eof), tag)
+            return None
+
+        if isinstance(source, io.StringIO):
+            input_bytes = source.getvalue().encode(*_split(opts.encoding))
+            self.add_task(redir.write_stream(input_bytes, stream, eof), tag)
+            return None
+
+        return stream
 
     def _setup_output_sink(self, stream, sink, encoding, tag):
         "Set up a task to write to custom output sink."
@@ -547,6 +619,9 @@ class Runner:
         elif isinstance(sink, Logger):
             self.add_task(redir.copy_logger(stream, sink, encoding), tag)
             stream = None
+        elif isinstance(sink, asyncio.StreamWriter):
+            self.add_task(redir.copy_streamwriter(stream, sink), tag)
+            stream = None
         return stream
 
     @log_method(LOG_EXIT, exc_value=2)
@@ -557,20 +632,22 @@ class Runner:
             suppress = await self._finish(exc_value)
         except asyncio.CancelledError:
             LOGGER.info("Runner cancelled inside _finish %r", self)
-            self.cancelled = True
+            self._cancelled = True
+        finally:
+            self._audit_callback("stop")
         return suppress
 
     @log_method(LOG_DETAIL)
     async def _finish(self, exc_value):
         "Finish the run. Return True only if `exc_value` should be suppressed."
-        assert self.proc
+        assert self._proc
 
         try:
             if exc_value is not None:
                 if _is_cancelled(exc_value):
-                    self.cancelled = True
+                    self._cancelled = True
                 await self._kill()
-                return self.cancelled
+                return self._cancelled
 
             await self._wait()
             return False
@@ -581,46 +658,55 @@ class Runner:
     @log_method(LOG_DETAIL)
     async def _close(self):
         "Make sure that our resources are properly closed."
-        assert self.proc
+        assert self._proc
 
-        if self.options.pty_fds:
-            self.options.pty_fds.close()
+        if self._options.pty_fds:
+            self._options.pty_fds.close()
 
         # _close can be called when unwinding exceptions. We need to handle
         # the case that the process has not exited yet. Remember to close the
         # transport.
-        if self.proc.returncode is None:
-            LOGGER.critical("Runner._close process still running %r", self.proc)
-            self.proc._transport.close()  # pylint: disable=protected-access
+        if self._proc.returncode is None:
+            LOGGER.critical("Runner._close process still running %r", self._proc)
+            self._proc._transport.close()  # pylint: disable=protected-access
             return
 
-        # asyncio child watcher artifact.
-        if self.proc.returncode == _FLAKY_EXIT_CODE:
-            LOGGER.warning("Runner._close exit code=%r", self.proc.returncode)
-
         try:
             # Make sure the transport is closed (for asyncio and uvloop).
-            self.proc._transport.close()  # pylint: disable=protected-access
+            self._proc._transport.close()  # pylint: disable=protected-access
 
             # Make sure that original stdin is properly closed. `wait_closed`
             # will raise a BrokenPipeError if not all input was properly written.
-            if self.proc.stdin is not None:
-                self.proc.stdin.close()
+            if self._proc.stdin is not None:
+                self._proc.stdin.close()
                 await harvest(
-                    self.proc.stdin.wait_closed(),
+                    self._proc.stdin.wait_closed(),
                     timeout=_CLOSE_TIMEOUT,
                     cancel_finish=True,  # finish `wait_closed` if cancelled
                     trustee=self,
                 )
 
         except asyncio.TimeoutError:
-            LOGGER.critical("Runner._close %r timeout stdin=%r", self, self.proc.stdin)
+            LOGGER.critical("Runner._close %r timeout stdin=%r", self, self._proc.stdin)
+
+    def _audit_callback(self, phase, *, failure=None, signal=None):
+        "Call `audit_callback` if there is one."
+        callback = self.command.options.audit_callback
+        if callback:
+            info = {"runner": self}
+            if failure:
+                info["failure"] = failure
+            if phase == "signal":
+                if signal is None:
+                    signal = "Signals.SIGKILL"  # SIGKILL (even on win32)
+                info["signal"] = str(signal)
+            callback(phase, info)
 
     def __repr__(self):
         "Return string representation of Runner."
-        cancelled = " cancelled" if self.cancelled else ""
-        if self.proc:
-            procinfo = f" pid={self.proc.pid} exit_code={self.proc.returncode}"
+        cancelled = " cancelled" if self._cancelled else ""
+        if self._proc:
+            procinfo = f" pid={self._proc.pid} exit_code={self._proc.returncode}"
         else:
             procinfo = " pid=None"
         return f"<Runner {self.name!r}{cancelled}{procinfo}>"
@@ -632,7 +718,7 @@ class Runner:
 
         stream = self.stdout or self.stderr
         if stream:
-            async for line in redir.read_lines(stream, self.options.encoding):
+            async for line in redir.read_lines(stream, self._options.encoding):
                 yield line
 
     def __aiter__(self):
@@ -673,55 +759,58 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
     ```
     """
 
+    stdin = None
+    "Pipeline standard input."
+
+    stdout = None
+    "Pipeline standard output."
+
+    stderr = None
+    "Pipeline standard error."
+
     def __init__(self, pipe, *, capturing):
         """`capturing=True` indicates we are within an `async with` block and
         client needs to access `stdin` and `stderr` streams.
         """
         assert len(pipe.commands) > 1
 
-        self.pipe = pipe
-        self.cancelled = False
-        self.tasks = []
-        self.results = None
-        self.capturing = capturing
-        self.encoding = pipe.options.encoding
-        self.stdin = None
-        self.stdout = None
-        self.stderr = None
+        self._pipe = pipe
+        self._cancelled = False
+        self._tasks = []
+        self._results = None
+        self._capturing = capturing
+        self._encoding = pipe.options.encoding
 
     @property
     def name(self):
         "Return name of the pipeline."
-        return self.pipe.name
+        return self._pipe.name
 
     def result(self):
         "Return `Result` object for PipeRunner."
-        return make_result(self.pipe, self.results, self.cancelled)
+        return make_result(self._pipe, self._results, self._cancelled)
 
-    def add_task(self, coro, tag=None):
+    def add_task(self, coro, tag=""):
         "Add a background task."
-        if tag:
-            task_name = f"{self.name}#{tag}"
-        else:
-            task_name = self.name
+        task_name = f"{self.name}#{tag}"
         task = asyncio.create_task(coro, name=task_name)
-        self.tasks.append(task)
+        self._tasks.append(task)
         return task
 
     @log_method(LOG_DETAIL)
     async def _wait(self, *, kill=False):
         "Wait for pipeline to finish."
-        assert self.results is None
-        assert self.tasks is not None
+        assert self._results is None
+        assert self._tasks is not None
 
         if kill:
             LOGGER.info("PipeRunner.wait killing pipe %r", self)
-            for task in self.tasks:
+            for task in self._tasks:
                 task.cancel()
 
-        cancelled, self.results = await harvest_results(*self.tasks, trustee=self)
+        cancelled, self._results = await harvest_results(*self._tasks, trustee=self)
         if cancelled:
-            self.cancelled = True
+            self._cancelled = True
 
     @log_method(LOG_ENTER)
     async def __aenter__(self):
@@ -731,7 +820,7 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
         except (Exception, asyncio.CancelledError) as ex:
             LOGGER.warning("PipeRunner enter %r ex=%r", self, ex)
             if _is_cancelled(ex):
-                self.cancelled = True
+                self._cancelled = True
             await self._wait(kill=True)
             raise
 
@@ -743,7 +832,7 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
             suppress = await self._finish(exc_value)
         except asyncio.CancelledError:
             LOGGER.warning("PipeRunner cancelled inside _finish %r", self)
-            self.cancelled = True
+            self._cancelled = True
         return suppress
 
     @log_method(LOG_DETAIL)
@@ -751,9 +840,9 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
         "Wait for pipeline to exit and handle cancellation."
         if exc_value is not None:
             if _is_cancelled(exc_value):
-                self.cancelled = True
+                self._cancelled = True
             await self._wait(kill=True)
-            return self.cancelled
+            return self._cancelled
 
         await self._wait()
 
@@ -769,7 +858,7 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
 
             cmds = self._setup_pipeline(open_fds)
 
-            if self.capturing:
+            if self._capturing:
                 stdin, stdout, stderr = await self._setup_capturing(cmds)
             else:
                 for cmd in cmds:
@@ -792,7 +881,7 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
         Each created open file descriptor is added to `open_fds` so it can
         be closed if there's an exception later.
         """
-        cmds = list(self.pipe.commands)
+        cmds = list(self._pipe.commands)
 
         cmd_count = len(cmds)
         for i in range(cmd_count - 1):
@@ -839,11 +928,11 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
     def __repr__(self):
         "Return string representation of PipeRunner."
         cancelled_info = ""
-        if self.cancelled:
+        if self._cancelled:
             cancelled_info = " cancelled"
         result_info = ""
-        if self.results:
-            result_info = f" results={self.results!r}"
+        if self._results:
+            result_info = f" results={self._results!r}"
         return f"<PipeRunner {self.name!r}{cancelled_info}{result_info}>"
 
     async def _readlines(self):
@@ -853,7 +942,7 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
 
         stream = self.stdout or self.stderr
         if stream:
-            async for line in redir.read_lines(stream, self.encoding):
+            async for line in redir.read_lines(stream, self._encoding):
                 yield line
 
     def __aiter__(self):
