@@ -1,11 +1,18 @@
 "Implements various utility functions."
 
+import asyncio
+import contextvars
 import io
 import os
+import shutil
+from collections import defaultdict
 from typing import Any, Iterable, Optional, Union
 
 from .log import LOG_DETAIL, LOGGER, log_timer
 
+# Stores current stack of context managers for immutable Command objects.
+_CTXT_STACK = contextvars.ContextVar("ctxt_stack", default=None)
+
 
 def decode(data: Optional[bytes], encoding: str) -> str:
     "Utility function to decode optional byte strings."
@@ -59,11 +66,14 @@ def verify_dev_fd(fdesc: int) -> None:
 def wait_pid(pid: int) -> Optional[int]:
     """Call os.waitpid and return exit status.
 
+    Not supported on Windows; raises an AttributeError.
+
     Return None if process is still running.
     """
     assert pid is not None and pid > 0
 
     try:
+        # os.WNOHANG is not available on Windows.
         result_pid, status = os.waitpid(pid, os.WNOHANG)
     except ChildProcessError as ex:
         # Set status to 255 if process not found.
@@ -79,8 +89,57 @@ def wait_pid(pid: int) -> Optional[int]:
     # Convert os.waitpid status to an exit status.
     try:
         status = os.waitstatus_to_exitcode(status)
-    except ValueError:
-        # See https://github.com/python/cpython/blob/7e5c107541726b90d3f2e6e69ef37180cf58335d/Lib/asyncio/unix_events.py#L47
+    except ValueError:  # pragma: no cover
+        # waitstatus_to_exitcode can theoretically raise a ValueError if
+        # the status is not understood. In this case, we do what
+        # asyncio/unix_events.py does: return the original status.
         pass
 
     return status
+
+
+async def uninterrupted(coro):
+    "Run a coroutine so it completes even if the current task is cancelled."
+
+    task = asyncio.create_task(coro)
+    try:
+        return await asyncio.shield(task)
+
+    except asyncio.CancelledError:
+        if not task.cancelled():
+            await task
+        raise
+
+
+def which(command):
+    "Given a command without a directory, return the fully qualified path."
+    path = shutil.which(command)
+    if path is None:
+        raise FileNotFoundError(command)
+    return path
+
+
+async def context_aenter(scope, ctxt_manager):
+    "Enter an async context manager."
+    ctxt_stack = _CTXT_STACK.get()
+    if ctxt_stack is None:
+        ctxt_stack = defaultdict(list)
+        _CTXT_STACK.set(ctxt_stack)
+
+    result = await ctxt_manager.__aenter__()
+    stack = ctxt_stack[scope]
+    stack.append(ctxt_manager)
+
+    return result
+
+
+async def context_aexit(scope, exc_type, exc_value, exc_tb):
+    "Exit an async context manager."
+    ctxt_stack = _CTXT_STACK.get()
+
+    stack = ctxt_stack[scope]
+    ctxt_manager = stack.pop()
+    if not stack:
+        _CTXT_STACK.set(None)
+
+    return await ctxt_manager.__aexit__(exc_type, exc_value, exc_tb)
